/* 
 *  resample~ by Volker Böhm
 *
 * sinc-interpolated variable speed sample-playback
 *  using libresample 0.1.3:
 *		Real-time library interface by Dominic Mazzoni
 *		Based on resample-1.7:
 *		http://www-ccrma.stanford.edu/~jos/resample/
 *
 * License: LGPL - see the file LICENSE.txt for more information
 */

#include "ext.h"
#include "z_dsp.h"
#include "buffer.h"	

#include "libresample.h"

void *resample_class;

typedef struct _resample
{
	t_pxobject l_obj;
	t_symbol *l_sym;
	t_buffer *l_buf;
	long	l_chan;
	long	outchns;				// object's number of output channels			
	double	ratio;					// magic gain factor 
	
	void	**handle;
	float	*srcc;
	float	**dst;
	int		step, numout, loop, stop;
	long	pos, ptr, frames;
} t_resample;

t_int *resample_perform1chn(t_int *w);
t_int *resample_perform2chns(t_int *w);
void resample_dsp(t_resample *x, t_signal **sp);
void resample_set(t_resample *x, t_symbol *s);
void *resample_new(t_symbol *s, long chan);
void resample_float(t_resample *x, double f);
void resample_assist(t_resample *x, void *b, long m, long a, char *s);
void resample_dblclick(t_resample *x);
void resample_free(t_resample *x);
void resample_loop(t_resample *x, long r);
void resample_stop(t_resample *x);
void resample_bang(t_resample *x);

t_symbol *ps_buffer;

void main(void)
{
	setup((t_messlist **)&resample_class, (method)resample_new, (method)resample_free, (short)sizeof(t_resample), 0L, 
		A_SYM, A_DEFLONG, 0);
	addmess((method)resample_dsp, "dsp", A_CANT, 0);
	addmess((method)resample_set, "set", A_SYM, 0);
	addmess((method)resample_loop, "loop", A_LONG, 0);
	addmess((method)resample_stop, "stop", 0);
	addbang((method)resample_bang);
	addfloat((method)resample_float);
	addmess((method)resample_assist, "assist", A_CANT, 0);
	addmess((method)resample_dblclick, "dblclick", A_CANT, 0);
	dsp_initclass();
	ps_buffer = gensym("buffer~");
	
	post("resample~ by vb\n\t---> using libresample by d.mazzoni\n\t---> based on resample-1.7 by JOS");
}


t_int *resample_perform1chn(t_int *w)
{
	t_resample *x = (t_resample *)(w[1]);
	t_float *in = (t_float *)(w[2]);
	t_float *out = (t_float *)(w[3]);
	int n = (int)(w[4]);
	t_buffer *b = x->l_buf;
	float *tab;
	float **dst;
	float *src;
	double temp, ratio;
	long chan,frames,nc, pos;
	unsigned long ptr;
	long saveinuse;
	int numout, dstlen, inUsed, i, loop;
	
	ptr = x->ptr;
	dst = x->dst;
	pos = x->pos;
	loop = x->loop;
	numout = x->numout;
	inUsed = 0;
	
	if (x->l_obj.z_disabled)
		goto out;
	if (!b)
		goto zero;
	if (!b->b_valid)
		goto zero;
	if(pos<0 || x->stop)
		goto zero;
	
	saveinuse = b->b_inuse;
	b->b_inuse = true;
	tab = b->b_samples;
	chan = x->l_chan;
	frames = x->frames = b->b_frames;
	nc = b->b_nchans;
	
	while (--n) {
		temp = *++in;
		
		if(ptr>=numout){
			if(temp<0.013) ratio = 80.0;
			else if(temp>10.0) ratio = 0.1;
			else ratio = 1.0/temp;
			
			dstlen = x->step * ratio + 1000;
			if(nc==1) {
				src = &tab[pos];
				numout = resample_process(x->handle[0], ratio, src, x->step, 0, &inUsed, dst[0], dstlen);
			}
			else if(nc==2) {
				for(i=0; i<x->step; i++)
					x->srcc[i] = tab[(i+pos)*nc];
				numout = resample_process(x->handle[0], ratio, x->srcc, x->step, 0, &inUsed, dst[0], dstlen);
			}
			else {
				*++out = 0.0;
				goto zeroplus;
			}
			
			ptr = 0;				//reset pointer to output buffer (dst)
			pos += inUsed;		// increment input pointer by number of 'used' samples
			//if(pos>frames-x->step) pos=0;		// wrap input pointer if end of buffer is reached
		}
		if(pos>frames-x->step) {
			if(loop) pos=0;
			else pos = -1;
		}
		
		*++out = dst[0][ptr++];	// read values from output buffer (dst)
	}
	x->numout = numout;
	x->pos = pos;
	x->ptr = ptr;
	
	b->b_inuse = saveinuse;
	return w + 5;
	
zeroplus:
		while (--n)
			*++out = 0.;
		b->b_inuse = saveinuse;
		return w + 5;
zero:
		while (--n)
			*++out = 0.;
out:
		return w + 5;
}



t_int *resample_perform2chns(t_int *w)
{
	t_resample *x = (t_resample *)(w[1]);
	t_float *in = (t_float *)(w[2]);
	t_float *out = (t_float *)(w[3]);
	t_float *out2 = (t_float *)(w[4]);
	int n = (int)(w[5]);
	t_buffer *b = x->l_buf;
	float *tab, *src;
	float **dst;
	double temp, ratio;
	long chan,frames,nc, pos, ende, saveinuse;
	unsigned long ptr;
	int numout, dstlen, inUsed, loop, c, i;
	
	ptr = x->ptr;
	dst = x->dst;
	pos = x->pos;
	loop = x->loop;
	numout = x->numout;
	inUsed = 0;
	
	if (x->l_obj.z_disabled)
		goto out;
	if (!b)
		goto zero;
	if (!b->b_valid)
		goto zero;
	if(pos<0 || x->stop)
		goto zero;
	
	saveinuse = b->b_inuse;
	b->b_inuse = true;
	tab = b->b_samples;
	chan = x->l_chan;
	frames = b->b_frames;
	nc = b->b_nchans;
	x->frames = frames;
	ende = frames - x->step;
	
	if(nc == 1) {					// mono buffer-------------------------
		while (--n) {
			temp = *++in;
			
			if(ptr>=numout){
				if(temp<0.013) ratio = 80.0;
				else if(temp>10.0) ratio = 0.1;
				else ratio = 1.0/temp;

				dstlen = x->step * ratio + 1000;
				src = &tab[pos];
				numout = resample_process(x->handle[0], ratio, src, x->step, 0, &inUsed, dst[0], dstlen);
				
				ptr = 0;				//reset pointer to output buffer (dst)
				pos += inUsed;		// increment input pointer by number of 'used' samples
				//if(pos>frames-x->step) pos=0;		// wrap input pointer if end of buffer is reached
			}
			if(pos>ende) {		// muss ich nach hier nehmen, wegen new soundfile->new frames
				if(loop) pos=0;	
				else pos = -1;
			}
			*++out = dst[0][ptr];	// read values from output buffer (dst)
			*++out2 = dst[0][ptr];
			ptr++;
		}
	}
	
	else if(nc==2) {				// stereo buffer--------------------------
		while (--n) {
			temp = *++in;
			/*
			if(temp<0.01 || temp>10.) {
				*++out = 0.0;
				*++out2 = 0.0;
				goto zeroplus;
			}*/
			
			if(ptr>=numout){
				if(temp<0.013) ratio = 80.0;
				else if(temp>10.0) ratio = 0.1;
				else ratio = 1.0/temp;
				
				dstlen = x->step * ratio + 1000;

				for(c=0; c<nc; c++) {
					for(i=0; i<x->step; i++)
						x->srcc[i] = tab[(i+pos)*nc+c];		// copy interleaved sample data channel by channel
					numout = resample_process(x->handle[c+1], ratio, x->srcc, x->step, 0, &inUsed, dst[c], dstlen);
				}

				ptr = 0;				//reset pointer to output buffer (dst)
				pos += inUsed;		// increment input pointer by number of 'used' samples
				//if(pos>frames-x->step) pos=0;		// wrap input pointer if end of buffer is reached
			}
			if(pos>ende) {		// muss ich nach hier nehmen, wegen new soundfile->new frames
				if(loop) pos=0;	
				else pos = -1;
			}
			*++out = dst[0][ptr];	// read values from output buffer (dst)
			*++out2 = dst[1][ptr];
			ptr++;
		}
	}
	
	else goto zero;		// only mono or stereo buffers supported

	
	x->numout = numout;
	x->pos = pos;
	x->ptr = ptr;
	
	b->b_inuse = saveinuse;
	return w + 6;
	
zeroplus:
	while (--n)
		*++out = *++out2 = 0.;
	b->b_inuse = saveinuse;
	return w + 6;
	
zero:
	while (--n)
		*++out = *++out2 = 0.;
out:
	return w + 6;
}




void resample_bang(t_resample *x) {
	resample_float(x, 0.0);
}


void resample_loop(t_resample *x, long r) {
	if( r == 0 ) x->loop= 0;
	else x->loop = 1;
} 


void resample_stop(t_resample *x) {
	x->stop = 1;
}

// here's where we set the buffer~ we're going to access
void resample_set(t_resample *x, t_symbol *s)
{
	t_buffer *b;
	
	x->l_sym = s;
	if ((b = (t_buffer *)(s->s_thing)) && ob_sym(b) == ps_buffer) {
		x->l_buf = b;
	} else {
		error("resample~: no buffer~ %s", s->s_name);
		x->l_buf = 0;
	}
}


void resample_float(t_resample *x, double f) {
	x->stop = 0;
	if(f>=0 &&f <= 1) 
		x->pos = f * (x->frames-x->step);
}



void resample_dsp(t_resample *x, t_signal **sp)
{
	resample_set(x,x->l_sym);
	if(x->outchns==1)
		dsp_add(resample_perform1chn, 4, x, sp[0]->s_vec-1, sp[1]->s_vec-1, sp[0]->s_n+1);
	else if(x->outchns ==2)
		dsp_add(resample_perform2chns, 5, x, sp[0]->s_vec-1, sp[1]->s_vec-1, sp[2]->s_vec-1, sp[0]->s_n+1);
}

// this lets us double-click on resample~ to open up the buffer~ it references
void resample_dblclick(t_resample *x)
{
	t_buffer *b;
	
	if ((b = (t_buffer *)(x->l_sym->s_thing)) && ob_sym(b) == ps_buffer)
		mess0((t_object *)b,gensym("dblclick"));
}


void resample_assist(t_resample *x, void *b, long m, long a, char *s)
{
	if (m == ASSIST_OUTLET)
		switch(a) {
			case 0:
				sprintf(s,"(signal) audio out channel 1");
				break;
			case 1:
				sprintf(s,"(signal) audio out channel 2");
				break;
		}
	else 
		sprintf(s,"(signal) playback speed, (float) playback pos (0.-1.)");

}




void resample_free(t_resample *x) {
	//int i;
	dsp_free((t_pxobject *)x);
	if(x->handle) {
		//sysmem_freeptr(x->handle[0]);
		//for(i=0; i<outchns
		free(x->handle[0]);
		free(x->handle[1]);
		free(x->handle[2]);
		free(x->dst[0]);
		free(x->dst[1]);
		
		free(x->handle);
		free(x->dst);
	}
	if(x->srcc) free(x->srcc);
}

void *resample_new(t_symbol *s, long outchns)
{
	int i, c;
	t_resample *x = (t_resample *)newobject(resample_class);
	dsp_setup((t_pxobject *)x, 1);
	//intin((t_object *)x,1);
	
	if(outchns==0) outchns=1;
	else if(outchns>2) outchns = 2;
	x->outchns = outchns;
	for(i=0; i<outchns; i++)
		outlet_new((t_object *)x, "signal");

	x->l_sym = s;


	x->ptr = 0;
	x->pos = 0;
	x->ratio = 0.8;
	x->numout = 0;
	x->frames = 0;
	x->step = 128;					// very small, OK ???
	x->loop = x->stop = 0;
	
	x->handle = (void **)malloc(3 * sizeof(void *));		// need three handles, wegen mono/stereo
	x->dst = (float **)malloc(2 * sizeof(float *));
	
	for(c=0; c<2; c++)
		x->dst[c] = (float *)malloc((x->step*100.0+1000)*sizeof(float));		//dstlen = step * ratio + 1000

	for(c=0; c<3; c++)
		x->handle[c] = resample_open(0, 0.1, 100.0);		// quality, min ratio, max ratio
	
	x->srcc = (float *)malloc(x->step * sizeof(float));
	return (x);
}

